#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2014, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

import laygo2.object.template
import laygo2.object.physical
import laygo2.object.database

import numpy as np
import yaml
import pprint

# Template library for the advanced example technology (advtech).

# Technology parameters
if __name__ == '__main__':
    tech_fname = './technology_example.yaml'
else:
    tech_fname = '../test_tech/technology_example.yaml'

with open(tech_fname, 'r') as stream:
    try:
        tech_params = yaml.safe_load(stream)
    except yaml.YAMLError as exc:
        print(exc)


# Primitive devices
def mos_bbox_func(devtype, params):
    """Computes x and y coordinate values from params."""
    unit_size = tech_params['templates']['mos'][devtype]['unit_size']
    xy0 = [0, 0]
    xy1 = [unit_size[0]*params['nf'], unit_size[1]]
    return np.array([xy0, xy1])


def pmos_bbox_func(params):
    return mos_bbox_func(devtype='pmos', params=params)


def nmos_bbox_func(params):
    return mos_bbox_func(devtype='nmos', params=params)


def mos_pins_func(devtype, params):
    """Generate a pin dictionary from params."""
    unit_size = tech_params['templates']['mos'][devtype]['unit_size']
    base_params = tech_params['templates']['mos'][devtype]['pins']
    nf = params['nf']
    sd_swap = params['sd_swap'] if 'sd_swap' in params.keys() else False
    pins = dict()
    lpin_default = 'S'
    rpin_default = 'D'
    if sd_swap:
        lpin = rpin_default
        rpin = lpin_default
    else:
        lpin = lpin_default
        rpin = rpin_default
    # pins starting from the left-most diffusion. Source if sd_swap=False, else Drain.
    _elem = []
    for fidx in range(nf // 2 + 1):
        l_xy = base_params[lpin_default]['xy'] + np.array([fidx, 0]) * unit_size*2
        l_layer = base_params[lpin_default]['layer']
        _elem.append(laygo2.object.Pin(xy=l_xy, layer=l_layer, netname=lpin))
    pins[lpin] = laygo2.object.Pin(xy=base_params[lpin_default]['xy'], layer=l_layer, netname=lpin, elements=np.array(_elem))
    # pins starting from the the other side of diffusion. Drain if sd_swap=False, else Source.
    _elem = []
    for fidx in range(((nf + 1) // 2)):
        r_xy = base_params[rpin_default]['xy'] + np.array([fidx, 0]) * unit_size*2
        r_layer = base_params[rpin_default]['layer']
        _elem.append(laygo2.object.Pin(xy=r_xy, layer=r_layer, netname=rpin))
    pins[rpin] = laygo2.object.Pin(xy=base_params[rpin_default]['xy'], layer=r_layer, netname=rpin, elements=np.array(_elem))
    # gate
    g_xy = np.array([base_params['G']['xy'][0], base_params['G']['xy'][1] + np.array([nf - 1, 0]) * unit_size])
    g_layer = base_params['G']['layer']
    pins['G'] = laygo2.object.Pin(xy=g_xy, layer=g_layer, netname='G')
    #g_xy = base_params['G']['xy'] + np.array([nf, 0]) * unit_size
    #_elem = []
    #for fidx in range(nf):
    #    g_xy = base_params['G']['xy'] + np.array([fidx, 0]) * unit_size
    #    g_layer = base_params['G']['layer']
    #    _elem.append(laygo2.object.Pin(xy=g_xy, layer=g_layer, netname='G'))
    #pins['G'] = laygo2.object.Pin(xy=base_params['G']['xy'], layer=g_layer, netname='G', elements=np.array(_elem))
    return pins


def pmos_pins_func(params):
    return mos_pins_func(devtype='pmos', params=params)


def nmos_pins_func(params):
    return mos_pins_func(devtype='nmos', params=params)


def mos_generate_func(devtype, name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
    """Generates an instance from the input parameters."""
    # Compute parameters
    nf = params['nf'] if 'nf' in params.keys() else 1
    unit_size = np.array(tech_params['templates']['mos'][devtype]['unit_size'])

    # Create the base mosfet structure.
    nelements = dict()
    for rn, rpar in tech_params['templates']['mos'][devtype]['rects'].items():
        # Computing loop index
        idx_start = 0 if 'idx_start' not in rpar else rpar['idx_start']
        idx_end = 0 if 'idx_end' not in rpar else rpar['idx_end']
        idx_step = 1 if 'idx_step' not in rpar else rpar['idx_step']
        for i in range(idx_start, nf+idx_end, idx_step):
            rxy = rpar['xy'] + np.array([i, 0]) * unit_size
            nelements[rn+str(i)] = laygo2.object.Rect(xy=rxy, layer=rpar['layer'], name=rn+str(i))
    #if 'rects_merged' in tech_params['templates']['mos'][devtype]:
    #    for rn, rpar in tech_params['templates']['mos'][devtype]['rects_merged'].items():
    #        rxy = np.array([[rpar['xy'][0][0], rpar['xy'][0][1]], [rpar['xy'][1][0] + unit_size[0] * (nf - 1), rpar['xy'][1][1]]])
    #        nelements[rn] = laygo2.object.Rect(xy=rxy, layer=rpar['layer'], name=rn)

    # Create pins
    pins = mos_pins_func(devtype=devtype, params=params)
    nelements.update(pins)  # Add physical pin structures to the virtual object.

    inst_unit_size = unit_size * np.array([nf, 1])
    # Generate and return the final instance
    inst = laygo2.object.VirtualInstance(name=name, xy=np.array([0, 0]), libname='mylib', cellname='myvcell_'+devtype,
                                         native_elements=nelements, shape=shape, pitch=pitch,
                                         transform=transform, unit_size=inst_unit_size, pins=pins)
    return inst


def pmos_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
    return mos_generate_func(devtype='pmos', name=name, shape=shape, pitch=pitch, transform=transform, params=params)


def nmos_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
    return mos_generate_func(devtype='nmos', name=name, shape=shape, pitch=pitch, transform=transform, params=params)


# Routing vias
def via_generate_func(devtype, name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
    """Generates an instance from the input parameters."""
    # Create the via structure.
    nelements = dict()
    for rn, rpar in tech_params['templates']['via'][devtype]['rects'].items():
        nelements[rn] = laygo2.object.Rect(xy=rpar['xy'], layer=rpar['layer'], name=rn)

    # Generate and return the final instance
    inst = laygo2.object.VirtualInstance(name=name, xy=np.array([0, 0]), libname='mylib', cellname='myvcell_devtype',
                                         native_elements=nelements, shape=shape, pitch=pitch,
                                         transform=transform, unit_size=np.array([0, 0]), pins=None)
    return inst


def via_r12_default_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
    return via_generate_func(devtype='via_r12_default', name=name, shape=shape, pitch=pitch, transform=transform,
                             params=params)


def via_r12_bottomplug_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
    return via_generate_func(devtype='via_r12_bottomplug', name=name, shape=shape, pitch=pitch, transform=transform,
                             params=params)


def via_r12_topplug_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
    return via_generate_func(devtype='via_r12_topplug', name=name, shape=shape, pitch=pitch, transform=transform,
                             params=params)


def via_r23_default_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
    return via_generate_func(devtype='via_r23_default', name=name, shape=shape, pitch=pitch, transform=transform,
                             params=params)


# Template library
def load_templates():
    """Load template to a template library object"""
    tlib = laygo2.object.database.TemplateLibrary(name='advtech_templates')
    # Transistors
    # Transistor layouts are created in laygo and stored as a virtual instance.
    tnmos = laygo2.object.template.UserDefinedTemplate(name='nmos', bbox_func=nmos_bbox_func, pins_func=nmos_pins_func,
                                                       generate_func=nmos_generate_func)
    tpmos = laygo2.object.template.UserDefinedTemplate(name='pmos', bbox_func=pmos_bbox_func, pins_func=pmos_pins_func,
                                                       generate_func=pmos_generate_func)
    tlib.append([tpmos, tnmos])
    # Vias
    # Via layouts are created in laygo and stored as a virtual instance.
    # tvia_r12_0 = laygo2.object.template.NativeInstanceTemplate(libname='advtech_templates', cellname='via_r12_0')
    # tlib.append([tvia_r12_0])
    tvia_r12_default = laygo2.object.template.UserDefinedTemplate(name='via_r12_default', bbox_func=lambda params: np.array([0, 0]),
                                                                  pins_func=lambda params: None,
                                                                  generate_func=via_r12_default_generate_func)
    tlib.append([tvia_r12_default])
    tvia_r12_topplug = laygo2.object.template.UserDefinedTemplate(name='via_r12_topplug', bbox_func=lambda params: np.array([0, 0]),
                                                                  pins_func=lambda params: None,
                                                                  generate_func=via_r12_topplug_generate_func)
    tlib.append([tvia_r12_topplug])
    tvia_r12_bottomplug = laygo2.object.template.UserDefinedTemplate(name='via_r12_bottomplug', bbox_func=lambda params: np.array([0, 0]),
                                                                     pins_func=lambda params: None,
                                                                     generate_func=via_r12_bottomplug_generate_func)
    tlib.append([tvia_r12_bottomplug])
    tvia_r23_default = laygo2.object.template.UserDefinedTemplate(name='via_r23_default', bbox_func=lambda params: np.array([0, 0]),
                                                                  pins_func=lambda params: None,
                                                                  generate_func=via_r23_default_generate_func)
    tlib.append([tvia_r23_default])
    return tlib


# Tests
if __name__ == '__main__':
    # Create templates.
    print("Create templates")
    templates = load_templates()
    print(templates['nmos'])
    print(templates['pmos'])



# -*- coding: utf-8 -*-
########################################################################################################################
#
# Copyright (c) 2014, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""laygo2 object package implements physical/template/grid/design-hierarchy objects."""

from laygo2.object import *
from laygo2.object.database import *
from laygo2.object.grid import *
from laygo2.object.physical import *
from laygo2.object.template import *




import pytest
import numpy as np
import pprint
import laygo2

import test_tech as tech

def test_grid_manual_convertion_table():
    g1   = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 100], elements=[0, 10, 20, 40, 50])
    g1_y = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 100], elements=[10, 20, 40, 50, 60])
    g2 = laygo2.object.grid.Grid(name='test', vgrid=g1, hgrid=g1_y)

    set_op1 = ("[]", "()", "==", "<", "<=", ">", ">=")

    g1ap, g1pa = g1.abs2phy, g1.phy2abs

    set_g1         = ( g1[10], "dummy" , g1 == 10 , g1 < 10 , g1 <=10 , g1 > 10 , g1 >= 10  )
    set_g1_abs2phy = ( g1ap[10], g1ap(10), g1ap ==10, g1ap < 10 , g1ap <=10, g1ap >10, g1ap >=10 )
    set_g1_phy2abs = ( g1pa[10], g1pa(10), g1pa ==10, g1pa < 10 , g1pa <=10, g1pa >10, g1pa >=10 )

    set_g1_title = ("g1", "g1.abs2phy", "g1.phy2abs")
    set_g1_loop  = (set_g1, set_g1_abs2phy, set_g1_phy2abs)

    g2x, g2y, g2xy, g2m, g2n, g2mn  = g2.x, g2.y, g2.xy, g2.m, g2.n, g2.mn

    set_g2    = ( g2[[10,10]]   , "dummy" , g2 == [ 10,10 ], g2 < [10,10 ] , g2 <= [10,10 ], g2 > [10,10 ], g2 >= [10,10 ] )
    set_g2_x  = ( g2x[10], g2x(10), g2x ==10 , g2x < 10 , g2x <= 10, g2x > 10 , g2x >= 10 )
    set_g2_y  = ( g2y[10], g2y(10), g2y ==10 , g2y < 10 , g2y <= 10, g2y > 10 , g2y >= 10 )
    set_g2_xy = ( g2xy[[10,10]] , g2xy( [10,10] )   , g2xy == [ 10,10 ], g2xy < [10,10 ] , g2xy <= [10,10 ], g2xy > [10,10 ], g2xy >= [10,10 ] )

    set_g2_m  = (g2m[10], g2m(10), g2m ==10 , g2m < 10 , g2m <= 10, g2m > 10 , g2m >= 10)
    set_g2_n  = (g2n[10], g2n(10), g2n ==10 , g2n < 10 , g2n <= 10, g2n > 10 , g2n >= 10)
    set_g2_mn = (g2mn[[10,10]] , g2mn( [10,10] )   , g2mn == [ 10,10 ], g2mn < [10,10 ] , g2mn <= [10,10 ], g2mn > [10,10 ], g2mn >= [10,10 ])

    set_loop  = ( set_g2, set_g2_x, set_g2_y, set_g2_xy, set_g2_m, set_g2_n, set_g2_mn, )
    set_title = (   "g2","g2_x","g2_y","g2_xy","g2_m", "g2_n", "g2_mn" )

    # g1
    for title, plist in zip(set_g1_title, set_g1_loop):
        print()
        print(title)
        print(set_op1)
        print(plist)
    # g2
    for title, plist in zip( set_title, set_loop ) :
        print( )
        print(title)
        print(set_op1)
        print(plist)

def test_grid_manual_Grid_conversion_operators():
    ## abs2phy  ##
    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 180], elements=[0, 35, 85, 130])
    g1_y = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 30], elements=[0])
    g2   = laygo2.object.grid.Grid(name='test', vgrid = g1_x, hgrid = g1_y)
    print("start")

    print("phy 2 abs ")
    test_title_phy2abs=[
        "__eq__", "__gt__", "__lt__", "__getitem__", "__call__", "__eq__", "__gt__", "__lt__"
    ]

    test_upper_phy2abs=[
        g2 == [35,0], g2 > [40, 0], g2 < [40, 0], g2.mn[35, 0], g2.mn([35, 0]), g2.xy == [35,0], g2.xy > [35,0], g2.xy < [35,0] ]

    test_lower_phy2abs = [
        g2 == [-95, 0], g2 > [-95, 0], g2 < [-95, 0], g2.mn[-95, 0], g2.mn([-95, 0]), g2.xy == [-95, 0], g2.xy > [-95, 0],
        g2.xy < [-95, 0]]


    test_title_abs2phy=[
       "__getitem__",  "__eq__", "__gt__", "__lt__", "__getitem__", "__call__"
    ]

    test_upper_abs2phy = [
        g2[1, 0], g2.mn == [1, 0], g2.mn > [1, 0], g2.mn < [1, 0],
        g2.xy[1, 0], g2.xy([1, 0]) ]

    test_lower_abs2phy = [
        g2[-1, 0], g2.mn == [-1, 0], g2.mn > [-1, 0], g2.mn < [-1, 0],
        g2.xy[-1, 0], g2.xy([-1, 0])]

    #for t, mn in zip( test_title, test_upper_phy2abs) :
    #    print( t, mn )

    #for t, mn in zip( test_title_phy2abs, test_lower_phy2abs) :
    #    print( t, mn )
    #for t, mn in zip( test_title_abs2phy, test_upper_abs2phy) :
    #    print( t, mn )

    for t, mn in zip( test_title_abs2phy, test_lower_abs2phy) :
        print( t, mn )

def test_grid_manual_CircularMapping():
    elements = [ 0, 35, 85, 130]
    cm = laygo2.object.CircularMapping( elements = elements)
    test_set = [
        [cm.dtype, "dtype"],
        [cm.elements, "elements"],
        [cm.shape, "shape"]
    ]

    for v, t in test_set:
        print(t, v, type(v))
    print(cm[5])
    print(cm[0:10])

def test_grid_manual_CircularMappingArray():
    elementx = [0, 35, 85, 130 ]
    elementy = [0 , 35]
    elements  = [[0,0], [35,0], [85,0], [130,0],
                 [35,35], [85, 35], [130,35 ]  ]
    cm = laygo2.object.CircularMapping(elements=elements)
    test_set = [
        [cm.dtype, "dtype"],
        [cm.elements, "elements"],
        [cm.shape, "shape"]
    ]
    print(elements)
    for v, t in test_set:
        print(t, v, type(v))
    print( cm[1] )
    print(cm[[0, 1]])

def test_grid_manual_AbsToPhyGridConverter():
    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 180], elements=[0, 35, 85, 130, 50])
    abs2phy = laygo2.object.grid._AbsToPhyGridConverter( master = g1_x)
    print( abs2phy)
    test_set=[
        [abs2phy.master, "master"],
        [abs2phy[0], "__getItem__ "],
        [abs2phy(0), " call"],
        [abs2phy == 35, " equal"],
        [abs2phy < 35, "lt"],
        [abs2phy <= 35, "te"],
        [abs2phy > 35, "gt"],
        [abs2phy >= 35, "ge"]
    ]

    for v, t in test_set:
        print( t, v, type(v) )

def test_grid_manual_PhyToAbsGridConverter():
    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 180], elements=[0, 35, 85, 130, 50])
    phy2abs = laygo2.object.grid._PhyToAbsGridConverter(master=g1_x)
    print(" ")
    print(phy2abs)
    test_set = [
        [phy2abs.master, "master"],
        [phy2abs[35], "__getItem__ "],
        [phy2abs(35), " call"],
        [phy2abs == 1, " equal"],
        [phy2abs < 1, "lt"],
        [phy2abs <= 1, "te"],
        [phy2abs > 1, "gt"],
        [phy2abs >= 1, "ge"],
    ]
    for v, t in test_set:
        print(t, v, type(v))

    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 100], elements=[10, 20, 40, 50, 60])
    g1_y = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 100], elements=[10, 20, 40, 50, 60])
    g2 = laygo2.object.grid.Grid(name='test', vgrid=g1_x, hgrid=g1_y)
    phy2abs = laygo2.object.grid._PhyToAbsGridConverter(master=g2)

    rect0 = laygo2.object.physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0')
    test_set2=[
      [phy2abs.bbox(rect0), "bbox"],
      [phy2abs.bottom_left(rect0), "bl"],
      [phy2abs.bottom_right(rect0), "br"],
      [phy2abs.top_left(rect0), "tl"],
      [phy2abs.top_right(rect0), "tr"],
      [phy2abs.center(rect0), "center"],
      [phy2abs.width(rect0), "width"],
      [phy2abs.height(rect0), "height"],
      [phy2abs.size(rect0), "size"],
    ]
    for t,v in test_set2:
        print(t,v, type(v))

    g1_y = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 120], elements=[0, 20, 40, 60, 80, 100, 120])
    g1_x = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 10], elements=[0])
    g2 = laygo2.object.grid.Grid(name='test', vgrid=g1_x, hgrid=g1_y)
    phy2abs=laygo2.object.grid._PhyToAbsGridConverter(master=g2)
    rect0 = laygo2.object.physical.Rect(xy=[[0, 0], [60, 90]], layer=['M1', 'drawing'], netname='net0',
                                        params={'maxI': 0.005})
    rect1 = laygo2.object.physical.Rect(xy=[[30, 30], [120, 120]], layer=['M1', 'drawing'], netname='net0',
                                        params={'maxI': 0.005})
    rect2 = laygo2.object.physical.Rect(xy=[[30, 30], [120, 120]], layer=['M1', 'drawing'], netname='net0',
                                        params={'maxI': 0.005})

    print("start")

    set_bbox2t = ("overlap", "crossing", "union ")
    set_bbox2 = (phy2abs.overlap(rect0, rect1), phy2abs.crossing(rect0, rect1), phy2abs.union(rect0, rect1))

    for t, v in zip(set_bbox2t, set_bbox2):
        print(t, v, type(v))

def test_grid_manual_OneDimGrid():
    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 180], elements=[0, 35, 85, 130, 50])
    print("  ")
    test_set=[
        [g1_x.name, "name"],
        [g1_x.range, "range"],
        [g1_x.phy2abs, "phy2abs"],
        [g1_x.abs2phy, "abs2phy"],
        [g1_x.width, "width"],
        [ g1_x.export_to_dict(), " export "]
    ]

    for v, t in test_set:
        print(t + "   ", v, type(v) )
    print(g1_x)

def test_grid_manual_Grid_xy():
    ## abs2phy  ##
    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 100], elements=[0, 10, 20, 40, 50])
    g1_y = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 100], elements=[10, 20, 40, 50, 60])
    g2   = laygo2.object.grid.Grid(name='test', vgrid = g1_x, hgrid = g1_y)

    set_op1 =( "[]", "()", "<", "<=", ">", ">=" )

    set_xy=(
        g2.xy[10, 10], g2.xy([10, 10]), g2.xy < [10,10], g2.xy <= [10, 10],
        g2.xy > [10,10], g2.xy >= [10, 10] )

    set_x = ( g2.x[10], g2.x([10]), g2.x < 10, g2.x <= 10, g2.x > 10, g2.x >= 10 )
    set_y = ( g2.y[10], g2.y([10]), g2.y < 10, g2.y <= 10, g2.y > 10, g2.y >= 10 )

    set_op2 = ("[]", "<", "<=", ">", ">=")
    set_g1x = ( g1_x[10],  g1_x < 10, g1_x <=10, g1_x >10 , g1_x >=10 )

    print("start")
    print(" 2-dimension, xy ")
    for token, value in zip(set_op1, set_xy):
        print(token + "  : ", value)


    print(" 2-dimension, x access ")
    for token, value in zip( set_op1, set_x):
        print( token + "  : ", value)

    print(" 1-dimension ")
    for token, value in zip( set_op2, set_g1x):
        print( token + "  : ", value)

def test_grid_manual_Grid_mn():
    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 100], elements=[0, 10, 20, 40, 50])
    g1_y = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 100], elements=[10, 20, 40, 50, 60])
    g2 = laygo2.object.grid.Grid(name='test', vgrid=g1_x, hgrid=g1_y)

    set_op1 = ("[]", "()", "<", "<=", ">", ">=")
    set_mn  = ( g2.mn[40, 40], g2.mn([40,40]), g2.mn < [40,40] , g2.mn <= [40, 40] , g2.mn > [40, 40] , g2.mn >= [40, 40]  )

    set_n = (g2.n[40], g2.n(40), g2.n < 40, g2.n <= 40, g2.n > 40, g2.n >= 40)
    print("start!")
    for token, value in zip( set_op1, set_mn):
        print(token + "  : ", value, "   :", type(value))

    print("n")
    for token, value in zip(set_op1, set_n):
        print(token + "  : ", value, "   :", type(value))

def test_grid_manual_Grid_property():
    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 100], elements=[0, 10, 20, 40, 50])
    g1_y = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 100], elements=[10, 20, 40, 50, 60])
    g2 = laygo2.object.grid.Grid(name='test', vgrid=g1_x, hgrid=g1_y)

    set_op1 = ("shape", "width", "height", "height_vec", "width_vec", "range")
    set_p  = ( g2.shape, g2.width, g2.height, g2.height_vec, g2.width_vec, g2.range  )

    for token, value in zip(set_op1, set_p):
        print(token + "  : ", value, "   :", type(value))

    print(g2.summarize())

def test_grid_manual_Grid_bboxHandler():
    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 100], elements=[10, 20, 40, 50, 60])
    g1_y = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 100], elements=[10, 20, 40, 50, 60])
    g2   = laygo2.object.grid.Grid(name='test', vgrid=g1_x, hgrid=g1_y)

    rect0 = laygo2.object.physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', params={'maxI': 0.005})
    rect1 = laygo2.object.physical.Rect(xy=[[50, 50], [150, 150]], layer=['M1', 'drawing'], netname='net0', params={'maxI': 0.005})
    rect2 = laygo2.object.physical.Rect(xy=[[70, 70], [150, 150]], layer=['M1', 'drawing'], netname='net0', params={'maxI': 0.005})

    pin0  = laygo2.object.physical.Pin(xy =[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', master=rect0,
               params={'direction': 'input'})

    #g2.bbox(pin0)
    set_bboxt= ( "bbox", "bl", "br", "tl", "tr")
    set_bbox = ( g2.bbox(pin0), g2.bottom_left(pin0), g2.bottom_right(pin0), g2.top_left(pin0), g2.top_right(pin0) )

    for t,v in zip(set_bboxt, set_bbox):
        print(t,v, type(v))

def test_grid_manual_Grid_bboxHandler2():
    g1_y = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 120], elements=[0,20,40,60,80,100,120])
    g1_x = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 10], elements=[0])
    g2 = laygo2.object.grid.Grid(name='test', vgrid=g1_x, hgrid=g1_y)

    rect0 = laygo2.object.physical.Rect(xy=[[0, 0], [ 60 , 90 ]], layer=['M1', 'drawing'], netname='net0',
                                        params={'maxI': 0.005})
    rect1 = laygo2.object.physical.Rect(xy=[[30, 30], [120, 120]], layer=['M1', 'drawing'], netname='net0',
                                        params={'maxI': 0.005})
    rect2 = laygo2.object.physical.Rect(xy=[[30, 30], [120, 120]], layer=['M1', 'drawing'], netname='net0',
                                        params={'maxI': 0.005})

    print("start")



    set_rectt   = ("rect0", "rect1", "rect2")
    set_rect_mn = (g2.mn( rect0.bbox ), g2.mn(rect1.bbox)  , g2.mn(rect2) )
    set_rect_xy = (rect0.bbox, rect1.bbox, rect2.bbox )

    for t, v, v2 in zip(set_rectt, set_rect_mn, set_rect_xy):
        print("\n", t, "\n",  v, "\n", v2)


    set_bbox2t = ("overlap", "crossing", "union ")
    set_bbox2 = (g2.overlap(rect0, rect1), g2.crossing(rect0, rect1), g2.union(rect0, rect1))

    for t, v in zip(set_bbox2t, set_bbox2):
        print(t, v, type(v))

def test_grid_manual_PlacementGrid():
    g1_x = laygo2.object.grid.OneDimGrid(name='xgrid', scope=[0, 20], elements=[0])
    g1_y = laygo2.object.grid.OneDimGrid(name='ygrid', scope=[0, 100], elements=[0])
    g2 = laygo2.object.grid.PlacementGrid(name='test', vgrid=g1_x, hgrid=g1_y)
    inst0_pins = dict()
    inst0_pins['in'] = laygo2.object.physical.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
    inst0_pins['out'] = laygo2.object.physical.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
    inst0 = laygo2.object.physical.Instance(name="I0", xy=[100, 100], libname="mylib", cellname="mycell"
                                            , shape=[3, 2], pitch=[200, 200], unit_size=[100, 100], pins=inst0_pins,
                                            transform='R0')

    print("start")
    print(inst0.xy)

    g2.place(inst=inst0, mn=[10,10])
    print(inst0.xy)

def test_grid_manual_RoutingGrid():
    templates = tech.load_templates()
    grids = tech.load_grids(templates=templates)
    pg, r12, r23 = grids['placement_cmos'], grids['routing_12_cmos'], grids['routing_23_cmos']


    print("start")
    print(r23.elements)
    test_list=[
        ("type"  , r23.type),
        ("vwidth", r23.vwidth),
        ("hwidth", r23.hwidth),
        ("vextension", r23.vextension),
        ("hextension", r23.hextension),
        ("vlayer", r23.vlayer),
        ("hlayer", r23.hlayer),
        ("pin_vlayer", r23.pin_vlayer),
        ("pin_hlayer", r23.pin_hlayer),
        ("viamap", r23.viamap),
        ("primary_grid", r23.primary_grid)
    ]
    for t, v in test_list:
        print(t, v, end="  ")
        print(type(v))

    mn_list = [[0, -2], [0, 1], [2, 1], [5, 1]]
    route = r23.route(mn=mn_list, via_tag=[True, None, True, True])
    print("route:")
    print(route)
    mn_list = [[0, -2], [1, 0], [2, 5]]
    via = r23.via(mn=mn_list)
    print("via:")
    print(via)
    mn_list = [[0, -2], [1, 0], [2, 5], [3, 4], [4, 5], [5, 5]]
    track = r23.route_via_track(mn=mn_list, track=[None, 0])
    print("track:")
    print(track)
    mn_list = [ [0, 0], [10, 10]]
    pin = r23.pin(mn=mn_list, name="pin")
    print("pin:")
    print(pin)

def test_physical_manual_PhysicalObject():

    physical = laygo2.object.physical.PhysicalObject( xy = [[0, 0], [200, 200]], name="test", params={'maxI': 0.005})
    print("start")
    set_titile1= ("name", "xy", "bbox", "master", "params")
    set_print=(physical.name, physical.xy, physical.bbox, physical.master, physical.params)


    set_title2  =("pointers", "left", "right", "top", "bottom", "center", "bottom_left", "bottom_right", "top_left", "top_right")
    set_pointer=(physical.pointers, physical.left, physical.right, physical.top, physical.bottom, physical.center, physical.bottom_left, physical.bottom_right,
                 physical.top_left, physical.top_right
                 )
    for t, x in zip(set_titile1, set_print):
        print( t  + "  " +  str(x)  + "  " + str(type(x))     )

    for t, x in zip(set_title2, set_pointer):
        print( t  + "  " +  str(x)  + "  " + str(type(x))     )

    print(physical)

def test_physical_manual_IterablePhysicalObject():
    phy0  = laygo2.object.physical.IterablePhysicalObject( xy=[[0, 0], [100, 100]], name="test" )
    phy1 = laygo2.object.physical.IterablePhysicalObject( xy=[[0, 0], [200, 200]], name="test")
    phy2 = laygo2.object.physical.IterablePhysicalObject( xy=[[0, 0], [300, 300]], name="test")
    element = [phy0, phy1, phy2]
    iphy0 = laygo2.object.physical.IterablePhysicalObject( xy=[[0, 0], [300, 300]], name="test", elements = element)

    test_set=[
        iphy0.elements,
        iphy0.shape,
        iphy0.ndenumerate()
    ]
    print(" ")
    for v in test_set:
        print(v)
        print(type(v))
    print(iphy0)

def test_physical_manual_instance():
    import numpy as np
    inst0_pins = dict()
    inst0_pins['in']  = laygo2.object.physical.Pin( xy = [[0, 0], [10,10]], layer = ['M1', 'drawing'], netname = 'in')
    inst0_pins['out'] = laygo2.object.physical.Pin( xy = [[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
    inst0 = laygo2.object.physical.Instance( name = "I0", xy = [ 100,100], libname="mylib", cellname="mycell"
    , shape = [3,2], pitch = [200,200], unit_size = [100, 100], pins = inst0_pins, transform =  'R0')


    test_value= [
        [inst0.libname,  "libname"],
        [inst0.cellname, "cellname"],
        [inst0.unit_size, "unit_size"],
        [inst0.transform, "transform"],
        [inst0.pins, "pins"],
        [inst0.xy0, "xy0"],
        [inst0.xy1, "xy1"],
        [inst0.size, "size"],
        [inst0.pitch, "pitch"],
        [inst0.spacing, " spacing "],
        [inst0.height, " height"],
        [inst0.width, "width"],
        [inst0.xy, "xy"],
        [inst0.bbox, "bbox"]
    ]

    for v,t  in test_value:
        print(t, v, end="  ")
        print(type(v))


    for idx, it in inst0.ndenumerate():
        print( "   ", idx, it)

    print("pins")
    print(inst0.pins["in"].shape)
    for idx, it in inst0.pins["in"].ndenumerate():
        print( "   ", idx, it)

    print( inst0[1,0].xy0 )
    print( inst0.pins["in"][1,1].xy)

def test_physical_manual_Rect():
    rect0 = laygo2.object.physical.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0', hextension=20, vextension=20)
    test_set=[
        [ rect0.layer, " layer"],
        [ rect0.netname, " netname"],
        [ rect0.hextension,  " hextension"],
        [ rect0.vextension,  " vextension"],
        [ rect0.height,  " height"],
        [ rect0.width,  " width"],
        [ rect0.size, " size"]
    ]
    for v, t in test_set:
        print(t, v, end="  ")
        print( type(v) )
    print(rect0)

def test_physical_manual_Path():
    path0 = laygo2.object.physical.Path(xy=[[0, 0], [0, 100]], width=10, extension=5, layer=['M1', 'drawing'], netname='net0')
    test_set = [
        [path0.layer, " layer"],
        [path0.netname, " netname"],
        [path0.extension, " extension"],
        [path0.width, " width"],
        ]
    for v,t  in test_set:
        print(t, v, end="  ")
        print(type(v))
    print(path0)

def test_physical_manual_Pin():
    pin0 = laygo2.object.physical.Pin( xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], netname='net0',
               params={'direction': 'input'})
    test_set=[
    [pin0.layer, " layer"],
    [pin0.netname, " netname"],
    [pin0.height, " height"],
    [pin0.width, " width"],
    [pin0.size, " size"],
    [pin0.bbox, " bbox"],
    [pin0.xy, " xy"]

    ]
    print(" ")
    for v, t in test_set:
        print(t, v, end="  ")
        print(type(v))

    print( pin0 )

def test_physical_manual_Text():
    text0 = laygo2.object.physical.Text(xy=[[ 0, 0], [100,100 ]], layer=['text', 'drawing'], text='test', params=None)
    test_set = [
        [text0.layer, " layer"],
        [text0.text, " text"],
    ]
    print(" ")
    for v, t in test_set:
        print(t, v, end="  ")
        print(type(v))

def test_physical_manual_VirtuialInstance():
    vinst0_pins = dict()
    vinst0_pins['in']  = laygo2.object.physical.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
    vinst0_pins['out'] = laygo2.object.physical.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
    vinst0_native_elements = dict()
    vinst0_native_elements['R0'] = laygo2.object.physical.Rect(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'])
    vinst0_native_elements['R1'] = laygo2.object.physical.Rect(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'])
    vinst0_native_elements['R2'] = laygo2.object.physical.Rect(xy=[[0, 0], [100, 100]], layer=['prBoundary', 'drawing'])
    vinst0 = laygo2.object.physical.VirtualInstance(name='I0', libname='mylib', cellname='myvcell', xy=[500, 500],
                            native_elements=vinst0_native_elements, shape=[3, 2], pitch=[100, 100],
                            unit_size=[100, 100], pins=vinst0_pins, transform='R0')

    test_set=[
        [vinst0.native_elements, "native_elements"]
    ]
    print("  ")
    for v, t in test_set:
        print(t, v, end="  ")
        print(type(v))

    print("  ")
    print("  ", vinst0)
    for idx, it in vinst0.ndenumerate():
        print("  ", idx, it)


    print(" pins")
    for idx, it in vinst0.pins['in'].ndenumerate():
        print("  ", idx, it)

def test_template_manual_NativeInstanceTemplate():
    print("NativeInstanceTemplate test")
    # define pins
    nat_temp_pins = dict()
    nat_temp_pins['in']  = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
    nat_temp_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
    # create a template
    nat_temp = laygo2.object.NativeInstanceTemplate(libname='mylib', cellname='mynattemplate', bbox=[[0, 0], [100, 100]],
                                      pins=nat_temp_pins)
    nat_inst = nat_temp.generate(name='mynatinst', shape=[2, 2], pitch=[100, 100], transform='R0')

    test_set = [
        [nat_temp.name,     " name"],
        [nat_temp.libname,  " libname"],
        [nat_temp.cellname, " cellname"],
        [nat_temp.height(), " heightname"],
        [nat_temp.width(),  " width"],
        [nat_temp.size(),   " size"],
        [nat_temp.bbox(), " bbox"],
        [nat_temp.pins(), " pins"],
        [nat_temp.generate(), " generatos"],
        [nat_temp.export_to_dict(), " ex"]
    ]

    for target, text in test_set:
        print(text, end= ":   ")
        print(target, type(target))

def test_template_manual_ParameterizedInstanceTemplate():
    print("ParameterizedInstanceTemplate test")
    pcell_temp_pins = dict()

    def pcell_bbox_func(params):
        if params==None:
            params={}
            params['W'] = 1
        return np.array([[0, 0], [100 , 100* params['W']]])

    # define the pin generation function.
    def pcell_pins_func(params):
        if params==None:
            params={"W":1}
        i = params['W']
        template_pins = dict()
        pin_in  = laygo2.object.Pin(xy =[ [ 0, 0], [100 , 0 ] ],      layer=['M1', 'pin'], netname='in')
        pin_out = laygo2.object.Pin(xy =[ [ 0, 100], [100 , 100* i]], layer=['M1', 'pin'], netname='out')
        template_pins['in' ] = pin_in
        template_pins['out'] = pin_out
        return template_pins

    # create a template
    pcell_temp = laygo2.object.ParameterizedInstanceTemplate(libname='mylib', cellname='mynattemplate', bbox_func=pcell_bbox_func,
                                      pins_func=pcell_pins_func)
    pcell_inst1 = pcell_temp.generate(name='I1', params={ "W":2, "L":1} )
    pcell_inst2 = pcell_temp.generate(name='I2', params={"W": 1, "L": 1})

    test_set = [
        [pcell_temp.name,     " name"],
        [pcell_temp.libname,  " libname"],
        [pcell_temp.cellname, " cellname"],
        [pcell_temp.height(), " heightname"],
        [pcell_temp.width(),  " width"],
        [pcell_temp.size(),   " size"],
        [pcell_temp.bbox(), " bbox"],
        [pcell_temp.pins(), " pins"],
        [pcell_inst1, " generatos"],
    ]

    for target, text in test_set:
        print(text, end= ":   ")
        print(target, type(target))

def test_template_manual_UserDefinedTemplate():
    print("UserDefinedTemplate test")

    # define the bbox computation function.
    def user_bbox_func(params):
        if params==None:
            params={}
            params['multi'] = 1
        return np.array([[0, 0], [100 * params['multi'], 100]])

    # define the pin generation function.
    def user_pins_func(params):
        if params==None:
            params={}
            params['multi'] = 1
        i = params['multi']
        template_pins = dict()
        pin_in  = laygo2.object.Pin(xy =[ [ 0, 0], [100 * i, 0 ] ],   layer=['M1', 'pin'], netname='in')
        pin_out = laygo2.object.Pin(xy=[ [ 0, 100], [100 * i, 100]], layer=['M1', 'pin'], netname='out')
        template_pins['in' ] = pin_in
        template_pins['out'] = pin_out
        return template_pins

    # define the instance generation function.
    def user_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
        if params==None:
            params={}
            params['multi'] = 1
        m = params['multi']
        shape = np.array([1, 1])
        inst_pins = user_pins_func(params)
        inst_native_elements = dict()

        inst_native_elements['left'] = laygo2.object.Rect(xy=[ [0, 0], [0,100]], layer=['M1', 'drawing'])
        ofst = np.array([100, 0])
        for i in range(m):
            bl  = np.array([0,0])
            tr  = np.array([100,100])
            inst_native_elements['center'+str(i)] = laygo2.object.Rect(xy=[ i*ofst + bl , i*ofst + tr ], layer=['M3', 'drawing'])

        inst_native_elements['right'] = laygo2.object.Rect(xy=[ m*ofst + [0, 0], m*ofst + [ 0 , 100]], layer=['M1', 'drawing'] )
        inst = laygo2.object.VirtualInstance(name=name, libname='mylib', cellname='myvinst', xy=np.array([0, 0]),
                                             native_elements=inst_native_elements, shape=shape,
                                             pitch=pitch, unit_size=[m * 100, 100], pins=inst_pins,
                                             transform=transform, params=params)
        return inst

    user_temp = laygo2.object.UserDefinedTemplate(name='myusertemplate', bbox_func=user_bbox_func, pins_func=user_pins_func,
                                    generate_func=user_generate_func)
    #user_inst = user_temp.generate(name='myuserinst',  params={'multi': 5})
    print(user_temp)
    test_set = [
        [user_temp.name, " name"],
        [user_temp.height(), " height"],
        [user_temp.width(), " width"],
        [user_temp.size(), " size"],
        [user_temp.bbox(), " bbox"],
        [user_temp.pins(), " pins"],
        [user_temp.generate(), " generatos"],

    ]

    for target, text in test_set:
        print(text, end=":   ")
        print(target, type(target))

    #print(user_inst)
    #for n,i in user_inst.native_elements.items():
    #    print(i)

def test_database_manual_BaseDatabase():
    base = laygo2.object.BaseDatabase(name='mycell')

    rect0 = laygo2.object.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], name='R0', netname='net0', params={'maxI': 0.005})
    rect1 = laygo2.object.Rect(xy=[[200, 0], [300, 100]], layer=['M1', 'drawing'], netname='net0', params={'maxI': 0.005})
    pin0 = laygo2.object.Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'pin'], netname='n0', master=rect0, params={'direction': 'input'})

    inst0_pins = dict()
    inst0_pins['in'] = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
    inst0_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
    inst0 = laygo2.object.Instance(name='I0', xy=[100, 100], libname='mylib', cellname='mycell', shape=[3, 2], pitch=[100, 100],
                     unit_size=[100, 100], pins=inst0_pins, transform='R0')

    base.append(rect0)
    base.append(rect1)
    base.append(pin0)
    base.append(inst0)

    p_name = [
    ( "name",base.name ),
    ("params", base.params),
    ("noname_index", base.noname_index),
    ("keys", base.keys()),
    ("elements", base.elements) ]

    print("start")
    for t, v in p_name:
        print(t, v, end="  ")
        print(type(v))

    print("items()")
    print(base.items())
    print(type(base.items()))

    print("__getItem__()")
    print(base["R0"])
    print("__setItem__()")
    rect2 = laygo2.object.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], name='R2', netname='net0')
    base["R2"] = rect2
    print("__iter__()")
    print(base.__iter__())
    print(type(base.__iter__()))
    for i in base:
        print(i)
    print("init")
    print(base)

def test_database_manual_library():
    lib = laygo2.object.Library(name='mylib')
    print(lib.name)
    print(lib)

def test_database_manual_design():
    dsn = laygo2.object.Design(name='dsn', libname="testlib")
    print("dsn:")
    print(dsn)
    rect0 = laygo2.object.Rect(xy=[[0, 0], [100, 100]], layer=['M1', 'drawing'], name='R0', netname='net0', params={'maxI': 0.005})
    pin0 = laygo2.object.Pin(xy=[[0, 0], [100, 100]], layer=['M1', 'pin'], netname='n0', master=rect0, params={'direction': 'input'})

    inst0_pins = dict()
    inst0_pins['in'] = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
    inst0_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
    inst0 = laygo2.object.Instance(name='I0', xy=[100, 100], libname='mylib', cellname='mycell', shape=[3, 2], pitch=[100, 100],
                     unit_size=[100, 100], pins=inst0_pins, transform='R0')

    vinst0_pins = dict()
    vinst0_pins['in'] = laygo2.object.physical.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
    vinst0_pins['out'] = laygo2.object.physical.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
    vinst0_native_elements = dict()
    vinst0_native_elements['R0'] = laygo2.object.physical.Rect(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'])
    vinst0 = laygo2.object.physical.VirtualInstance(name='VI0', libname='mylib', cellname='myvcell', xy=[500, 500],
                                                    native_elements=vinst0_native_elements, shape=[3, 2],
                                                    pitch=[100, 100],
                                                    unit_size=[100, 100], pins=vinst0_pins, transform='R0')

    text0 = laygo2.object.physical.Text(xy=[[ 0, 0], [100,100 ]], layer=['text', 'drawing'], text='test', params=None)

    dsn.append(rect0)
    dsn.append(pin0)
    dsn.append(inst0)
    dsn.append(vinst0)
    dsn.append(text0)


    p_name = [
        ("libname",dsn.libname ),
        ("cellname", dsn.cellname),
        ("rects", dsn.rects),
        ("pins", dsn.pins),
        ("texts", dsn.texts),
        ("instnaces", dsn.instances),
        ("virtual_instnaces", dsn.virtual_instances)
        ]
    print("start")

    for t, v in p_name:
        print(t, v, end="  ")
        print(type(v))

    print("get by layer")
    print( dsn.get_matchedrects_by_layer(['M1', 'drawing']))
    print("export to template")
    print(dsn.export_to_template())
    ## route, route_via_track,



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2014, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements interface with virtuoso in skill language

"""
import logging
from math import log10
from decimal import *

import numpy as np
import laygo2.object
import laygo2.util.transform as tf

__author__ = ""
__maintainer__ = ""
__status__ = "Prototype"


def _py2skill_number(value, scale=0.001):
    fmt_str = "%." + "%d" % (-1 * log10(scale) + 1) + "f "  # for truncations
    return fmt_str % (value * scale)


def _py2skill_list(pylist, scale=0.001):
    """Convert a python list object to a skill list."""
    list_str = "list( "
    for item in pylist:
        if isinstance(item, list):  # nested list
            list_str += _py2skill_list(item) + " "
        elif isinstance(item, np.ndarray):  # nested list
            list_str += _py2skill_list(item, scale=scale) + " "
        elif isinstance(item, str):
            list_str += "\"" + str(item) + "\" "
        elif isinstance(item, int) or isinstance(item, np.integer):
            # fmt_str = "%."+"%d" % (-1*log10(scale)+1)+"f "  # for truncations
            # list_str += fmt_str%(item*scale) + " "
            list_str += _py2skill_number(item, scale) + " "
    list_str += ")"
    return list_str


def _translate_obj(objname, obj, scale=0.001, master=None, offset=np.array([0, 0])):
    """
    Convert an object to corresponding scale commands.
    offset : np.array([int, int])
        Offsets to obj.xy
    """
    if master is None:  
        mxy = np.array([0, 0])
        mtf = 'R0'
    else: # if the translated object has a master (e.g. VirtualInstance)
        mxy = master.xy
        mtf = master.transform
    if obj.__class__ == laygo2.object.Rect:
        color = obj.color # coloring func. added
        # Invoke _laygo2_generate_rect( cv layer bbox ) in {header_filename}
        _xy = np.sort(obj.xy, axis=0)  # make sure obj.xy is sorted
        _xy = mxy + np.dot(_xy + np.array([[-obj.hextension, -obj.vextension], [obj.hextension, obj.vextension]]),
                           tf.Mt(mtf).T)
        #_xy = mxy + np.dot(obj.xy + np.array([[-obj.hextension, -obj.vextension], [obj.hextension, obj.vextension]]),
        #                   tf.Mt(mtf).T)
        return "_laygo2_generate_rect(cv, %s, %s, \"%s\") ; for the Rect object %s \n" \
               % (_py2skill_list(obj.layer), _py2skill_list(_xy, scale=scale), color, objname) # coloring func. added
    elif obj.__class__ == laygo2.object.Path:
        # TODO: implement path export function.
        pass
    elif obj.__class__ == laygo2.object.Pin:
        if obj.elements is None:
            _objelem = [obj]
        else:
            _objelem = obj.elements
        for idx, _obj in np.ndenumerate(_objelem):
            # Invoke _laygo2_generate_pin(cv, name, layer, bbox) in {header_filename}
            _xy = mxy + np.dot(_obj.xy, tf.Mt(mtf).T)
            return "_laygo2_generate_pin(cv, \"%s\", %s, %s ) ; for the Pin object %s \n" \
                   % (_obj.netname, _py2skill_list(_obj.layer), _py2skill_list(_xy, scale=scale),
                      objname)
    elif obj.__class__ == laygo2.object.Text:
        # TODO: implement text export function.
        pass
    elif obj.__class__ == laygo2.object.Instance:
        # Invoke _laygo2_generate_instance( cv name libname cellname viewname loc orient num_rows num_cols
        # sp_rows sp_cols params params_order )
        _xy = mxy + np.dot(obj.xy, tf.Mt(mtf).T)
        if master is None:  
            transform = obj.transform
        else: # if the translated object has a master (e.g. VirtualInstance)
            transform = tf.combine(obj.transform, master.transform)
        if obj.shape is None:
            num_rows = 1
            num_cols = 1
            sp_rows = 0
            sp_cols = 0
        else:
            num_rows = obj.shape[1]
            num_cols = obj.shape[0]
            sp_rows = _py2skill_number(obj.pitch[1])
            sp_cols = _py2skill_number(obj.pitch[0])
        return "_laygo2_generate_instance(cv, \"%s\", \"%s\", \"%s\", \"%s\", %s, \"%s\", %d, %d, %s, %s, %s, %s) " \
               "; for the Instance object %s \n" \
               % (objname, obj.libname, obj.cellname, obj.viewname, _py2skill_list(_xy), transform,
                  num_rows, num_cols, sp_rows, sp_cols, "nil", "nil", objname)
    elif obj.__class__ == laygo2.object.VirtualInstance:
        cmd = ""
        if obj.shape is None:
            for elem_name, elem in obj.native_elements.items():
                if not elem.__class__ == laygo2.object.Pin:
                    if obj.name == None:
                        obj.name='NoName'
                    else:
                        pass
                    cmd += _translate_obj(obj.name + '_' + elem_name, elem, scale=scale, master=obj)
        else:  # arrayed VirtualInstance
            for i, j in np.ndindex(tuple(obj.shape.tolist())):  # iterate over obj.shape
                for elem_name, elem in obj.native_elements.items():
                    if not elem.__class__ == laygo2.object.Pin:
                        cmd += _translate_obj(obj.name + '_' + elem_name + str(i) + '_' + str(j), 
                                              elem, scale=scale, master=obj[i, j])            
        return cmd
    else:
        return obj.translate_to_skill()  #

    return ""

def export(db, filename=None, cellname=None, scale=1e-3, reset_library=False, tech_library=None):
    """
    Export a design object to skill code.

    Parameters
    ----------
    db: laygo2.database.Library
        Design to export
    filename: str, optional
        If specified, the generated skill script is stored in filename.
    cellname: str or List[str]
        The name(s) of cell(s) to be exported.
    scale: float
        The scaling factor between laygo2's integer coordinats actual physical coordinates.
    reset_library: bool, optional
        If True, the library to export the cells is reset.
    tech_library: str, optional
        The name of technology library to be attached to the resetted library.
    """
    # parse header functions.
    cmd = "; laygo2 layout export skill script.\n\n"
    import os
    header_filename = os.path.abspath(laygo2.interface.__file__)[:-11] + 'skill_export.il'
    with open(header_filename, 'r') as f:
        cmd += f.read()
        cmd += '\n'

    cellname = db.keys() if cellname is None else cellname  # export all cells if cellname is not given.
    cellname = [cellname] if isinstance(cellname, str) else cellname  # convert to a list for iteration.
    if reset_library:
        cmd += "_laygo2_reset_library(\"%s\" \"%s\")\n" % (db.name, tech_library)
    for cn in cellname:
        cmd += "\n; exporting %s__%s\n" % (db.name, cn)  # open the design.
        logging.debug('Export_to_SKILL: Cellname:' + cn)
        #cmd += "let( (cv) \n"  # limit the scope of cv. Disabled to prevent overflow.
        cmd += "cv = _laygo2_open_layout(\"%s\" \"%s\" \"layout\")\n" % (db.name, cn)  # open the design.
        # export objects
        for objname, obj in db[cn].items():
            cmd += _translate_obj(objname, obj, scale=scale)
        cmd += "_laygo2_save_and_close_layout(cv)\n"  # close the layout
        #cmd += ");let\n" Disabled to prevent overflow.
    if filename is not None:  # export to a file.
        with open(filename, "w") as f:
            f.write(cmd)
    return cmd



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2014, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements classes for various layout template objects.
"""

__author__ = "Jaeduk Han"
__maintainer__ = "Jaeduk Han"
__status__ = "Prototype"

from abc import *
import numpy as np
import laygo2.object


class Template(metaclass=ABCMeta):
    """
    The base class that defines the functions and attributes of the template.

    Attributes
    ----------
    name : str

    Methods
    -------
    __init__() 
    width()
    size() 
    bbox() 
    pins() 
    generate() 

    Notes
    -----
    Reference in Korean:
    Template의 기본동작과 속성을 정의한 기본 클래스.
    """

    name = None
    """str: Template name."""

    def __init__(self, name=None):
        """
        Constructor function of Template class.

        Parameters
        ----------
        name : str or None, optional.
            The name of this template.
        """
        self.name = name

    def __str__(self):
        """Return a string corresponding to this object's information."""
        return self.summarize()

    def summarize(self):
        """Return the summary of the template information."""
        return self.__repr__() + " " \
                                 "name: " + self.name + ", " + \
                                 "class: " + self.__class__.__name__ + ", " + \
                                 ""

    def height(self, params=None):
        """int: Return the height of a template."""
        return abs(self.bbox(params=params)[0, 1] - self.bbox(params=params)[1, 1])

    def width(self, params=None):
        """int: Return the width of a template."""
        return abs(self.bbox(params=params)[0, 0] - self.bbox(params=params)[1, 0])

    def size(self, params=None):
        """int: Return the size of a template."""
        return np.array([self.width(params=params), self.height(params=params)])

    @abstractmethod
    def bbox(self, params=None):
        """numpy.ndarray: (Abstract method) Return the bounding box of a template."""
        pass

    @abstractmethod
    def pins(self, params=None):
        """dict: (Abstract method) Return dict having the collection of pins of a template."""
        pass

    @abstractmethod
    def generate(self, name=None, shape=None, pitch=None, transform='R0', params=None):
        """instance: (Abstract method) Return the instance generated from a template."""
        pass


class NativeInstanceTemplate(Template):
    """
    NativeInstanceTemplate class implements the template that generate Instance.

    Attributes
    ----------
    name : str
    libname : str
    cellname : str

    Methods
    -------
    __init__() 
    height() 
    width() 
    size() 
    bbox() 
    pins() 
    generate() 
    export_to_dict() 

    Notes
    -----
    Reference in Korean:
    NativeInstanceTemplate 클래스는 기본 Instance를 반환하는 템플릿을 구현한다.
    """
    libname = None
    """attribute
    str: Library name of NativeInstanceTemplate object.

    

    Examples
    --------
    >>> nat_temp_pins = dict() >>> nat_temp_pins['in'] = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing’], netname='in’)
    >>> nat_temp_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing’], netname='out’)
    >>> nat_temp = laygo2.object.NativeInstanceTemplate(libname='mylib', cellname='mynattemplate’, bbox=[[0, 0], [100, 100]], pins=nat_temp_pins)
    >>> nat_temp.libname 
    “mylib”

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_NativeInstanceTemplate_libname.png

    Reference in Korean:
    str: NativeInstanceTemplate 객체의 library 이름.
    """

    cellname = None
    """attribute
    str: Cellname of NativeInstanceTemplate object.

    

    Examples
    --------
    >>> nat_temp_pins = dict() >>> nat_temp_pins['in']  = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing’], netname='in’)
    >>> nat_temp_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing’], netname='out’)
    >>> nat_temp = laygo2.object.NativeInstanceTemplate(libname='mylib', cellname='mynattemplate’, bbox=[[0, 0], [100, 100]], pins=nat_temp_pins)
    >>> nat_temp.cellname 
    “mynattemplate”

    Notes
    -----
    Related Images:
    https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_NativeInstanceTemplate_cellname.png

    Reference in Korean:
    str: NativeInstanceTemplate 객체의 cellname.
    """

    _bbox = np.array([[0, 0], [0, 0]])
    
    _pins = None

    def __init__(self, libname, cellname, bbox=np.array([[0, 0], [0, 0]]), pins=None):
        """
        Constructor function of NativeInstanceTemplate class.

        Parameters
        ----------
        libname : str
            library name.
        cellname : str
            cell name.
        bbox : numpy.ndarray
            bbox.
        pins : dict
            dictionary having the pin object.

        Returns
        -------
        NativeInstanceTemplate

        

        Examples
        --------
        >>> nat_temp_pins = dict() 
        >>> nat_temp_pins['in']  = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing’], netname='in’)
        >>> nat_temp_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing’], netname='out’)
        >>> nat_temp = laygo2.object.NativeInstanceTemplate(libname='mylib', cellname='mynattemplate’, bbox=[[0, 0], [100, 100]], pins=nat_temp_pins)
        <laygo2.object.template.NativeInstanceTemplate object>

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_NativeInstanceTemplate_init.png

        Reference in Korean:
        NativeInstanceTemplate 클래스의 생성자함수.
        파라미터
        libname(str): library 이름
        cellname(str): cell 이름
        bbox(numpy.ndarray): bbox
        pins(dict): pin 객체를 갖고있는 dictionary
        반환값
        laygo2.NativeInstanceTemplate
        참조
        없음
        """
        self.libname = libname
        self.cellname = cellname
        self._bbox = None if bbox is None else np.asarray(bbox)
        self._pins = pins
        Template.__init__(self, name=cellname)

    def summarize(self):
        """Return the summary of the template information."""
        return self.__repr__() + " " \
                                 "name: " + self.name + ", " + \
               "class: " + self.__class__.__name__ + ", " + \
               "bbox: " + str(self.bbox().tolist()) + ", " + \
               "pins: " + str(self.pins()) + ", " + \
               ""

    # Core template functions
    def bbox(self, params=None):
        """
        bbox of NativeInstanceTemplate object.

        Parameters
        ----------
        None

        Returns
        -------
        numpy.ndarray

        

        Examples
        --------
        >>> nat_temp_pins = dict() 
        >>> nat_temp_pins['in']  = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing’], netname='in’)
        >>> nat_temp_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing’], netname='out’)
        >>> nat_temp = laygo2.object.NativeInstanceTemplate(libname='mylib', cellname='mynattemplate’, bbox=[[0, 0], [100, 100]], pins=nat_temp_pins)
        >>> nat_temp.bbox() 
        [[0,0], [100, 100]]

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_NativeInstanceTemplate_bbox.png

        Reference in Korean:
        NativeInstanceTemplate 객체의 bbox.
        파라미터
        없음
        반환값
        numpy.ndarray
        참조
        없음
        """
        return self._bbox

    def pins(self, params=None):
        """
        Return pin dictionary of NativeInstanceTemplate object.

        Parameters
        ----------
        None

        Returns
        -------
        dict

        

        Examples
        --------
        >>> nat_temp_pins = dict() 
        >>> nat_temp_pins['in']  = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing’], netname='in’)
        >>> nat_temp_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing’], netname='out’)
        >>> nat_temp = laygo2.object.NativeInstanceTemplate(libname='mylib', cellname='mynattemplate’, bbox=[[0, 0], [100, 100]], pins=nat_temp_pins)
        >>> nat_temp.pins() 
        {'in': <laygo2.object.physical.Pin object>, 'out': <laygo2.object.physical.Pin object>}

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_ParameterizedInstanceTemplate_pins.png

        Reference in Korean:
        NativeInstanceTemplate 객체의 pin dictionary 반환.
        파라미터
        없음
        반환값
        dict
        참조
        없음
        """
        return self._pins

    def generate(self, name=None, shape=None, pitch=None, transform='R0', params=None):
        """
        Generate Instance object.

        Parameters
        ----------
        name : str 
            name of the instance to be generated.
        shape : numpy.ndarray, optional.
            shape of the object to be generated.
        pitch : numpy.ndarray, optional.
            pitch of the object to be generated.
        params : dict, optional.
            dictionary having the object attributes.
        transform : str
            transformation attribute of the object to be generated.
        
        Returns
        -------
        Instance

        See Also
        --------
        Class Instance

        Examples
        --------
        >>> nat_temp_pins = dict()
        >>> nat_temp_pins['in'] = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
        >>> nat_temp_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
        >>> nat_temp = laygo2.object.NativeInstanceTemplate(libname='mylib', cellname='mynattemplate', bbox=[[0, 0], [100, 100]], pins=nat_temp_pins)
        >>> nat_temp.generate(name="I1")
        <laygo2.object.physical.Instance object>
        >>> nat_temp.generate(name="I2")
        <laygo2.object.physical.Instance object>

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_ParameterizedInstanceTemplate_generate.png

        Reference in Korean:
        Instance 객체 생성.
        파라미터
        name(str): 생성할 인스턴스의 이름
        shape(numpy.ndarray): 생성할 객체의 shape [ optional ]
        pitch(numpy.ndarray): 생성할 객체간의 간격 [ optional ]
        params(dict) : 객체의 속성을 갖는 Dictionary [ optional ]
        transform(str): 생성할 객체의 변환 속성 [ optional ]
        반환값
        laygo2.Instance: 생성된 개체
        참조
        Class Instance
        """
        return laygo2.object.physical.Instance(libname=self.libname, cellname=self.cellname, xy=np.array([0, 0]),
                                               shape=shape, pitch=pitch, unit_size=self.size(params), pins=self.pins(params),
                                               transform=transform, name=name, params=params)

    # I/O functions
    def export_to_dict(self):
        """
        Return Dictionary containing the information of NativeInstanceTemplate.

        Parameters
        ----------
        None

        Returns
        -------
        dict

        See Also
        --------
        Class Instance

        Examples
        --------
        >>> nat_temp_pins = dict()
        >>> nat_temp_pins['in'] = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
        >>> nat_temp_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
        >>> nat_temp = laygo2.object.NativeInstanceTemplate(libname='mylib', cellname='mynattemplate', bbox=[[0, 0], [100, 100]], pins=nat_temp_pins)
        >>> nat_temp.export_to_dict()
        {'libname': 'mylib', 'cellname': 'mynattemplate', 'bbox': [[0, 0], [100, 100]], 'pins': {'in': {'xy': [[0, 0], [10, 10]], 'layer': ['M1', 'drawing'], 'name': None, 'netname': 'in'}, 'out': {'xy': [[90, 9 0], [100, 100]], 'layer': ['M1', 'drawing'], 'name': None, 'netname': 'out'}}}

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_NativeInstanceTemplate_export_to_dict.png

        Reference in Korean:
        NativeInstanceTemplate 의 정보가 담긴 Dictonary 반환.
        파라미터
        없음
        반환값
        dict
        참조
        Class Instance
        """
        db = dict()
        db['libname'] = self.libname
        db['cellname'] = self.cellname
        db['bbox'] = self.bbox().tolist()
        db['pins'] = dict()
        for pn, p in self.pins().items():
            db['pins'][pn] = p.export_to_dict()
        return db


class ParameterizedInstanceTemplate(Template):
    """
    ParameterizedInstanceTemplate class implements the template that generate ParameterizedInstnace.

    Attributes
    ----------
    name : str
    libname : str
    cellname : str

    Methods
    -------
    __init__() 
    height() 
    width() 
    size() 
    bbox() 
    pins() 
    generate()

    Notes
    -----
    Reference in Korean:
    ParameterizedInstanceTemplate 클래스는 Parameterized Instance를 반환하는 템플릿을 구현한다.
    """

    libname = None
    """str: Libname of the instance being generated."""

    cellname = None
    """str: Cellname of the instance being generated."""

    _bbox = None

    _pins = None

    def __init__(self, libname, cellname, bbox_func=None, pins_func=None):
        """
        Generate ParameterizedInstanceTemplate object.

        Parameters
        ----------
        libname : str
            library name.
        cellname : str 
            The cell name of the template. 
        bbox_func : callable
            bbox.
        pins_func : callable
            dictionary having the pin object.
        
        Returns
        -------
        laygo2.NativeInstanceTemplate

        

        Examples
        --------
        >>> def pcell_bbox_func(params): 
            …… 
        >>> def pcell_pins_func(params): 
            ……
        >>> pcell_temp = laygo2.object.ParameterizedInstanceTemplate(libname='mylib', cellname='mynattemplate’, bbox_func= pcell_bbox_func, pins_func= pcell_pins_func)
        <laygo2.object.template.ParameterizedInstanceTemplate object>
        
        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_ParameterizedInstanceTemplate_init.png

        Reference in Korean:
        ParameterizedInstanceTemplate 클래스의 생성자함수.
        파라미터
        libname(str): library 이름 
        cellname(str): cell 이름 
        bbox_func(callable): bbox 
        pins_func(callable): pin 객체를 갖고있는 dictionary
        반환값
        laygo2.NativeInstanceTemplate
        참조
        없음
        """
        self.libname = libname
        self.cellname = cellname
        self._bbox = bbox_func
        self._pins = pins_func
        Template.__init__(self, name=cellname)

    # Core template functions
    def bbox(self, params=None):
        """
        bbox of ParameterizedInstanceTemplate object.

        Parameters
        ----------
        None

        Returns
        -------
        numpy.ndarray

        

        Examples
        --------
        >>> def pcell_bbox_func(params):
                if params==None: 
                    params={“W”:1}
                return np.array([[0, 0], [100 , 100* params['W']]])
        >>> def pcell_pins_func(params): 
            ……
        >>> pcell_temp = laygo2.object.ParameterizedInstanceTemplate(libname='mylib', cellname='mynattemplate’, bbox_func= pcell_bbox_func, pins_func= pcell_pins_func)
        >>> pcell_temp.bbox 
        [[0,0], [100,100]]

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_ParameterizedInstanceTemplate_bbox.png

        Reference in Korean:
        ParameterizedInstanceTemplate 객체의 bbox.
        파라미터
        없음
        반환값
        numpy.ndarray
        참조
        없음
        """
        return self._bbox(params=params)

    def pins(self, params=None):
        """
        Return pin dictionary of ParameterizedInstanceTemplate object.

        Parameters
        ----------
        None

        Returns
        -------
        dict

        

        Examples
        --------
        >>> def pcell_bbox_func(params): 
            ……
        >>> def pcell_pins_func(params):
                if params==None:
                    params={"W":1} 
                    i = params['W']
                template_pins = dict()
                pin_in  = laygo2.object.Pin(xy =[ [ 0, 0], [100 , 0 ] ],      layer=['M1', 'pin'], netname='in')
                pin_out = laygo2.object.Pin(xy =[ [ 0, 100], [100 , 100* i]], layer=['M1', 'pin’], netname='out')
                template_pins['in' ] = pin_in
                template_pins['out'] = pin_out
                return template_pins
        >>> pcell_temp = laygo2.object.ParameterizedInstanceTemplate(libname='mylib', cellname='mynattemplate’, bbox_func= pcell_bbox_func, pins_func= pcell_pins_func)
        >>> pcell_temp.pins 
        {'in': <laygo2.object.physical.Pin object>, 'out': <laygo2.object.physical.Pin object>}

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_ParameterizedInstanceTemplate_pins.png

        Reference in Korean:
        ParameterizedInstanceTemplate 객체의 pin dictionary 반환.
        파라미터
        없음
        반환값
        dict
        참조
        없음
        """
        return self._pins(params=params)

    def generate(self, name=None, shape=None, pitch=None, transform='R0', params=None):
        """
        Generate ParameterizedInstance object.

        Parameters
        ----------
        name : str
            name of the instance to be generated.
        shape : numpy.ndarray, optional.
            shape of the object to be generated.
        pitch : numpy.ndarray, optional.
            pitch of the object to be generated.
        params : dict, optional.
            dictionary having the entity attributes.
        transform : str, optional.
            transformation attribute of the entity to be generated.
        
        Returns
        -------
        (laygo2.Instance) generated object

        See Also
        --------
        Class Instance

        Examples
        --------
        >>> def pcell_bbox_func(params):
            ……
        >>> def pins_bbox_func(params): 
            ……
        >>> pcell_temp = laygo2.object.ParameterizedInstanceTemplate(libname='mylib', cellname='mynattemplate’, bbox_func=pcell_bbox_func, pins_func=pcell_pins_func)
        >>> pcell_temp.generate(name=“I1”, params={“W”=2, “L”=1}) 
        <laygo2.object.physical.Instance object>
        >>> pcell_temp.generate(name=“I2”, params={“W”=2, “L”=1}) 
        <laygo2.object.physical.Instance object>

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_ParameterizedInstanceTemplate_generate.png

        Reference in Korean:
        ParameterizedInstance 객체 생성.
        파라미터
        name(str): 생성할 인스턴스의 이름
        shape(numpy.ndarray): 생성할 객체의 shape [ optional ]
        pitch(numpy.ndarray): 생성할 객체간의 간격 [ optional ]
        params(dict) : 개체의 속성을 갖는 Dictionary [ optional ]
        transform(str):  생성할 개체의 변환 속성 [ optional ]
        반환값
        laygo2.Instance: 생성된 객체
        참조
        Class Instance
        """
        #xy = xy + np.dot(self.xy(params)[0], tf.Mt(transform).T)
        return laygo2.object.physical.Instance(libname=self.libname, cellname=self.cellname, xy=np.array([0, 0]),
                                               shape=shape, pitch=pitch, unit_size=self.size(params),
                                               pins=self.pins(params), transform=transform, name=name, params=params)


class UserDefinedTemplate(Template):
    """
    UserDefinedTemplate class implements the template that generate VirtualInstance.

    Attributes
    ----------
    name : str

    Methods
    -------
    __init__() 
    height() 
    width() 
    size() 
    bbox() 
    pins() 
    generate() 

    Notes
    -----
    Reference in Korean:
    UserDefinedTemplate 클래스는 VirtualInstance를 반환하는 템플릿을 구현한다.
    """

    _bbox = None

    _pins = None

    _generate = None

    def __init__(self, bbox_func, pins_func, generate_func, name=None):
        """
        Constructor function of UserDefinedTemplate class.

        Parameters
        ----------
        bbox_func: callable
            method computing bbox.
        pins_func: callable
            method computing pins.
        generate_func: callable
            method generating VirtualInstance.
        name : str
            template name.

        Returns
        -------
        laygo2.UserDefinedTemplate

        

        Examples
        --------
        >>> def user_bbox_func(params):-> numpy.ndarray …  ## return bbox0 * multi 
            …… 
        >>> def user_pins_func(params):-> dict          …  ## pin0.bbox = pins0.bbox * multi 
            …… 
        >>> def user_generate_func(params): 
            …… 
        >>> user_temp = laygo2.object.UserDefinedTemplate(name='myusertemplate', bbox_func=user_bbox_func, pins_func=user_pins_func, generate_func=user_generate_func)
        <laygo2.object.template.UserDefinedTemplate object>

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_UserDefinedTemplate_init.png

        Reference in Korean:
        UserDefinedTemplate 클래스의 생성자함수.
        파라미터
        bbox_func(callable ): bbox를 연산해주는 메소드
        pins_func(callable ): pins를 연산해주는 메소드
        generate_func(callable ): VirtualInstance를 생성하는 메소드
        name(str): 템플릿 이름
        반환값
        laygo2.UserDefinedTemplate
        참조
        없음
        """
        self._bbox = bbox_func
        self._pins = pins_func
        self._generate = generate_func
        Template.__init__(self, name=name)

    # Core template functions
    def bbox(self, params=None):
        """
        Return bbox of UserDefinedTemplate object.

        Parameters
        ----------
        None

        Returns
        -------
        numpy.ndarray

        

        Examples
        --------
        >>> params={}; params[“multi”] = 10; bbox0 = [ [0,0],[100,100]]; pin0 = [ in, out ] 
        >>> def user_bbox_func(params):-> numpy.ndarray …  ## return bbox0 * multi
                if params==None: 
                    params={}
                    params['multi'] = 1
                return np.array([[0, 0], [100 * params['multi'], 100]])
        >>> def user_pins_func(params):-> dict          …  ## pin0.bbox = pins0.bbox * multi 
            …… 
        >>> def user_generate_func(params): 
            …… 
        >>> user_temp = laygo2.object.UserDefinedTemplate(name='myusertemplate', bbox_func=user_bbox_func, pins_func=user_pins_func, generate_func=user_generate_func)
        >>> user_temp.bbox() 
        [[0, 0], [100, 100]]

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_UserDefinedTemplate_bbox.png

        Reference in Korean:
        UserDefinedTemplate 객체의 bbox 반환.
        파라미터
        없음
        반환값
        numpy.ndarray
        참조
        없음
        """
        return self._bbox(params=params)

    def pins(self, params=None):
        """
        Pins of UserDefinedTemplate object.

        Parameters
        ----------
        None

        Returns
        -------
        dict

        

        Examples
        --------
        >>> params={}; params[“multi”] = 10; bbox0 = [ [0,0],[100,100]]; pin0 = [ in, out ] 
        >>> def user_bbox_func(params):-> numpy.ndarray …  ## return bbox0 * multi 
        >>> def user_pins_func(params):-> dict
                if params==None:
                    params={"multi":1}
                i = params['multi']
                template_pins = dict()
                pin_in = laygo2.object.Pin(xy =[ [ 0, 0], [100 * i, 0 ] ],   layer=['M1', 'pin'], netname='in') 
                pin_out = laygo2.object.Pin(xy=[ [ 0, 100], [100 * i, 100]], layer=['M1', 'pin'], netname='out') 
                template_pins['in' ] = pin_in 
                template_pins['out'] = pin_out 
                return template_pins
        >>> def user_generate_func(params):
            …… 
        >>> user_temp = laygo2.object.UserDefinedTemplate(name='myusertemplate', bbox_func=user_bbox_func, pins_func=user_pins_func, generate_func=user_generate_func)
        >>> user_temp.pins()
        {'in': <laygo2.object.physical.Pin>, 'out': <laygo2.object.physical.Pin object>}

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_UserDefinedTemplate_pins.png

        Reference in Korean:
        UserDefinedTemplate 객체의 pins.
        파라미터
        없음
        반환값
        dict
        참조
        없음
        """
        return self._pins(params=params)

    def generate(self, name=None, shape=None, pitch=None, transform='R0', params=None):
        """
        Generate VirtualInstance object.

        Parameters
        ----------
        None

        Returns
        -------
        laygo2.VirtualInstance

        

        Examples
        --------
        >>> def user_bbox_func(params):-> numpy.ndarray …  ## return bbox0 * multi 
        >>> def user_pins_func(params):-> dict 
        >>> def user_generate_func(params={“multi”:1}):
                m = params['multi'] 
                shape = np.array([1, 1]); inst_pins = user_pins_func(params) ; inst_native_elements = dict() 
                inst_native_elements['left'] = laygo2.object.Rect(xy=[ [0, 0], [0,100]], layer=['M1’, 'drawing']) 
                ofst = np.array([100, 0])
                for i in range(m):
                    bl  = np.array([0,0]) 
                    tr  = np.array([100,100]) 
                    inst_native_elements['center'+str(i)] = laygo2.object.Rect(xy=[ i*ofst + bl , i*ofst + tr ]…
                inst_native_elements['right'] = laygo2.object.Rect(xy=[ m*ofst + [0, 0], m*ofst + [ 0 , 100]], …) 
                inst = VirtualInstance(name=name,xy=np.array([0, 0]),native_elements=inst_native_elements,……)
        >>> user_temp = laygo2.object.UserDefinedTemplate(name='myusertemplate', bbox_func=user_bbox_func, pins_func=user_pins_func, generate_func=user_generate_func) 
        >>> nat_temp.generate(name=“I1”, {“multi”=1}) 
        <laygo2.object.physical.VirtualInstance >
        >>> nat_temp.generate(name=“I2”, {“multi”=2}) 
        <laygo2.object.physical.VirtualInstance >

        Notes
        -----
        Related Images:
        https://github.com/niftylab/laygo2/tree/master/docs_workspace/assets/img/object_template_UserDefinedTemplate_generate.png

        Reference in Korean:
        VirtualInstance 객체 생성.
        파라미터
        없음
        반환값
        laygo2.VirtualInstance
        참조
        없음
        """
        return self._generate(name=name, shape=shape, pitch=pitch, transform=transform, params=params)


# Test
if __name__ == '__main__':
    test_native_template = True
    test_pcell_template = True
    test_user_template = True

    import laygo2.object

    if test_native_template:
        print("NativeInstanceTemplate test")
        # define pins
        nat_temp_pins = dict()
        nat_temp_pins['in'] = laygo2.object.Pin(xy=[[0, 0], [10, 10]], layer=['M1', 'drawing'], netname='in')
        nat_temp_pins['out'] = laygo2.object.Pin(xy=[[90, 90], [100, 100]], layer=['M1', 'drawing'], netname='out')
        # create a template
        nat_temp = NativeInstanceTemplate(libname='mylib', cellname='mynattemplate', bbox=[[0, 0], [100, 100]],
                                          pins=nat_temp_pins)
        # generate
        nat_inst = nat_temp.generate(name='mynatinst', shape=[2, 2], pitch=[100, 100], transform='R0')
        # display
        print(nat_temp)
        print(nat_inst)

    if test_pcell_template:
        print("ParameterizedInstanceTemplate test")

        # define the bbox computation function.
        def pcell_bbox_func(params):
            return np.array([[0, 0], [100 * params['mult'], 100]])

        # define the pin generation function.
        def pcell_pins_func(params):
            template_pins = dict()
            for i in range(params['mult']):
                template_pins['in' + str(i)] = laygo2.object.Pin(xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
                                                                 layer=['M1', 'drawing'], netname='in' + str(i))
                template_pins['out' + str(i)] = laygo2.object.Pin(xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
                                                                  layer=['M1', 'drawing'], netname='out' + str(i))
            return template_pins

        # create a template.
        pcell_temp = ParameterizedInstanceTemplate(libname='mylib', cellname='mypcelltemplate',
                                                   bbox_func=pcell_bbox_func, pins_func=pcell_pins_func)
        # generate based on the parameter assigned.
        pcell_inst_params = {'mult': 4}
        pcell_inst_size = pcell_temp.size(params=pcell_inst_params)
        pcell_inst = pcell_temp.generate(name='mypcellinst', shape=[2, 2], pitch=pcell_inst_size, transform='R0',
                                         params=pcell_inst_params)
        # display
        print(pcell_temp)
        print(pcell_inst)

    if test_user_template:
        print("UserDefinedTemplate test")

        # define the bbox computation function.
        def user_bbox_func(params):
            return np.array([[0, 0], [100 * params['mult'], 100]])

        # define the pin generation function.
        def user_pins_func(params):
            template_pins = dict()
            for i in range(params['mult']):
                template_pins['in' + str(i)] = laygo2.object.Pin(xy=[[i * 100 + 0, 0], [i * 100 + 10, 10]],
                                                                 layer=['M1', 'drawing'], netname='in' + str(i))
                template_pins['out' + str(i)] = laygo2.object.Pin(xy=[[i * 100 + 90, 90], [i * 100 + 90, 100]],
                                                                  layer=['M1', 'drawing'], netname='out' + str(i))
            return template_pins

        # define the instance generation function.
        def user_generate_func(name=None, shape=None, pitch=np.array([0, 0]), transform='R0', params=None):
            m = params['mult']
            shape = np.array([1, 1]) if shape is None else np.asarray(shape)

            inst_pins = user_pins_func(params)
            inst_native_elements = dict()
            for i in range(m):
                ofst = i * 100
                inst_native_elements['R0_' + str(i)] = laygo2.object.Rect(xy=[[ofst, 0], [ofst + 10, 10]],
                                                                          layer=['M1', 'drawing'])
                inst_native_elements['R1_' + str(i)] = laygo2.object.Rect(xy=[[ofst + 90, 90], [ofst + 100, 100]],
                                                                          layer=['M1', 'drawing'])
            inst_native_elements['R2'] = laygo2.object.Rect(xy=[[0, 0], [m * 100, 100]],
                                                            layer=['prBoundary', 'drawing'])
            inst = laygo2.object.VirtualInstance(name=name, libname='mylib', cellname='myvinst', xy=np.array([0, 0]),
                                                 native_elements=inst_native_elements, shape=shape,
                                                 pitch=pitch, unit_size=[m * 100, 100], pins=inst_pins,
                                                 transform=transform, params=params)
            return inst

        user_temp = UserDefinedTemplate(name='myusertemplate', bbox_func=user_bbox_func, pins_func=user_pins_func,
                                        generate_func=user_generate_func)
        user_inst = user_temp.generate(name='myuserinst', shape=[2, 1], params={'mult': 5})
        print(user_temp)
        print(user_inst)
        print(user_inst.bbox)



#!/usr/bin/python
########################################################################################################################
#
# Copyright (c) 2014, Regents of the University of California
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the
# following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following
#   disclaimer.
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the
#    following disclaimer in the documentation and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
# WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
########################################################################################################################

"""
This module implements interfaces with yaml files.

"""

__author__ = "Jaeduk Han"
__maintainer__ = "Jaeduk Han"
__status__ = "Prototype"

import yaml
import os.path
import laygo2

def export_template(template, filename, mode='append'):
    libname = template.libname
    cellname = template.cellname
    pins = template.pins()

    db = dict()
    if mode == 'append':  # in append mode, the template is appended to 'filename' if the file exists.
        if os.path.exists(filename):
            with open(filename, 'r') as stream:
                db = yaml.load(stream, Loader=yaml.FullLoader)
    if libname not in db:
        db[libname] = dict()
    db[libname][cellname] = template.export_to_dict()
    with open(filename, 'w') as stream:
        yaml.dump(db, stream)
    print("Your design was translated into YAML format.")

#filename=libname+'_templates.yaml'

def import_template(filename):
    # load yaml file
    if os.path.exists(filename):
        with open(filename, 'r') as stream:
            db = yaml.load(stream, Loader=yaml.FullLoader)
    libname = list(db.keys())[0]  # assuming there's only one library defined in each file.
    # create template library
    tlib = laygo2.object.database.TemplateLibrary(name=libname)
    # read out the yaml file entries and build template objects
    for tn, tdict in db[libname].items():
        pins = dict()
        if 'pins' in tdict:
            for pinname, pdict in tdict['pins'].items():
                pins[pinname] = laygo2.object.Pin(xy=pdict['xy'], layer=pdict['layer'], netname=pdict['netname'])
        t = laygo2.object.NativeInstanceTemplate(libname=libname, cellname=tn, bbox=tdict['bbox'], pins=pins)
        tlib.append(t)
    return tlib

